/*
 **************************************************************************
 *		SWI-Prolog ActiveX Extensions.
 *
 *	This software is provided in source and binary form free of charge to 
 *  users of SWI-Prolog.  No warranty of any kind is expressed or implied;
 *  in particular, this is entirely unsupported software.
 *
 *  Documentation for this product is available from the SWI-Prolog 
 *  web site (http://www.swi.psy.uva.nl/projects/SWI-Prolog/).
 *
 *	Version 1.0.1.
 *
 *	October 1, 1999
 *
 **************************************************************************
 */

//
// PrologQuery.cpp: Implementation of CPrologQuery
//
#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "declcom.h"
#include "proqcvt.h"

#include "ProQuery.h" 
#include "PrologQuery.h"

typedef vector<LONG> VLONG;

DECL_IP_U(IDispatch,IID_IDispatch);

//  Boolean atoms
static atom_t			atomTrue;				//  'true'
static atom_t			atomFalse;				//  'false'

/////////////////////////////////////////////////////////////////////////////
// CPrologQuery


CPrologQuery :: CHDIR :: CHDIR ( TSZC tszcPath )
{
	assert( tszcPath );
	TCHAR tchDir [ MAX_PATH ];
	DWORD dwLength = ::GetCurrentDirectory( sizeof tchDir - 1, tchDir );
	tchDir[dwLength] = 0;
	_zsPath = tchDir;
	bool bOK = ::SetCurrentDirectory( tszcPath );
	assert( bOK );
}

CPrologQuery :: CHDIR :: ~ CHDIR ()
{
	bool bOK = ::SetCurrentDirectory( _zsPath );
	assert( bOK );
}

void CPrologQuery :: InitProlog ()
{
	static bool bInited = false;
	if ( bInited )
		return;

	//  Construct a surrogate argument list for SWI Prolog.  Pass
	//		the name of this DLL as the executable.

	HINSTANCE hMod = _Module.m_hInst;
	TCHAR tchFn[MAX_PATH];
	DWORD dwLength = ::GetModuleFileName(hMod, tchFn, sizeof tchFn - 1 );
	if ( dwLength < 1 )
		return;
	tchFn[dwLength] = 0;
	char * argv[] = { tchFn, NULL };
	bInited = PL_initialise(1,argv);
	if ( ! bInited )
		return;

	//  Capture the current drive:path
	TCHAR tchDrive[MAX_PATH];
	TCHAR tchPath[MAX_PATH];
	_splitpath( tchFn, tchDrive, tchPath, NULL, NULL );	
	_zsModulePath.Format( "%s%s", tchDrive, tchPath );

	atomTrue	= PL_new_atom("true");
	atomFalse	= PL_new_atom("false");
}

STDMETHODIMP CPrologQuery::ModulePath(BSTR *pbstrPath)
{
	if ( _zsModulePath.length() == 0 )
		return E_FAIL;
	ZSTRW zswPath( _zsModulePath );
	_bstr_t bstrPath( zswPath.Tszc() );
	*pbstrPath = bstrPath.copy();
	return S_OK;
}


#define  NOTHANDLE  ULONG_MAX

int CPrologQuery :: _cInstances = 0;

CPrologQuery :: CPrologQuery()
	: _pred(NULL),
	_qid(NOTHANDLE),
	_psfaArgs(NULL),
	_tArgs(NULL),
	_iArity(0),
	_fframe(NOTHANDLE)
{
	InitProlog();
	_fframe = PL_open_foreign_frame();
	_cInstances++;
}

CPrologQuery ::	~ CPrologQuery()
{
	ResetAll();
	if ( _fframe != NOTHANDLE )
		PL_discard_foreign_frame(_fframe);
	_cInstances--;
}

HRESULT CPrologQuery :: FinalConstruct ()
{
	if ( _cInstances != 1 )
		return Error( IDS_ERR_NOT_ORIGINAL_INSTANCE );
	return S_OK;
}

HRESULT CPrologQuery :: Error ( UINT nID )
{
    return CComCoClass<CPrologQuery, &CLSID_PrologQuery>::Error(nID);
}

HRESULT CPrologQuery :: Error ( TSZC tszcDesc, const IID & iid, HRESULT hRes )
{
    ZSTRW zswDesc( tszcDesc );
    return CComCoClass<CPrologQuery, &CLSID_PrologQuery>::Error( zswDesc.Tszc(), iid, hRes);
}

void CPrologQuery :: ResetAll ()
{
	ResetQuery();
	_pred = NULL;
}

void CPrologQuery :: ResetQuery ()
{
	if ( _psfaArgs )
		::SafeArrayDestroy( _psfaArgs );
	_psfaArgs = NULL;

	if ( _qid != NOTHANDLE )
		PL_close_query( _qid );

	_qid = NOTHANDLE;
	_tArgs = NULL;
}

STDMETHODIMP CPrologQuery::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IPrologQuery
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

HRESULT CPrologQuery :: HrConvertTermsToSafeArray ( term_t tTerms, int cTerms, SAFEARRAY * * ppsfa )
{
	SAFEARRAYBOUND rgsabound[1];
	rgsabound[0].lLbound = 0;
	rgsabound[0].cElements = cTerms;

	HRESULT hr = S_OK;
	*ppsfa = NULL;
	SAFEARRAY * psfa = NULL;

	double dbl;
	int i;
	ZSTRW zsw;
	do
	{
		psfa = ::SafeArrayCreate( VT_VARIANT, 1, rgsabound );
		if ( ! psfa )
		{
			hr = Error( IDS_ERR_SFA_FAILURE );
			break;
		}
		hr = ::SafeArrayAllocData( psfa );
		if ( FAILED(hr) )
			break;
		
		for ( long iTerm = 0; iTerm < cTerms; iTerm++, tTerms++ )
		{			
			int iType = PL_term_type( tTerms );
			int iIsString = PL_is_string( tTerms );
			int iIsList = PL_is_list( tTerms );

			_variant_t vntTerm;

			switch ( iType )
			{
				case PL_TERM: //BUGBUG!  convert lists to SAFEARRAYs
				{
					if ( PL_is_list( tTerms ) )
					{
						if ( ! bConvertListToVariantSafearray( tTerms, vntTerm ) )
							hr = Error( IDS_ERR_INV_PLIST_CVT );
						break;
					}
				}
				case PL_STRING:
				case PL_ATOM:
				{
					char * psz;
					if ( ! PL_get_chars( tTerms, & psz, CVT_ALL ) )
					{
						hr = Error( IDS_ERR_PL_DATA_FAILURE );
						break;
					}
					vntTerm = psz;
					break;
				}

				case PL_INTEGER:
					PL_get_integer( tTerms, & i );
					vntTerm = (long)i;
					break;

				case PL_FLOAT:
					PL_get_float( tTerms, & dbl );
					vntTerm = dbl;
					break;

				case PL_VARIABLE:
					//  Unbound variables are returned as VT_EMPTY
					vntTerm.vt = VT_EMPTY;
					break;

				default:
					assert(false);
					hr = Error( IDS_ERR_UNEX_TERM_TYPE );
					break;
			}
			if ( FAILED(hr) )
				break;
			long iIndex = rgsabound[0].lLbound + iTerm;
			hr = ::SafeArrayPutElement( psfa, & iIndex, & vntTerm );
		}
	} 
	while (false);

	if ( FAILED(hr) )
	{
		if ( psfa )
			::SafeArrayDestroy( psfa );
		psfa = NULL;
	}
	else
	{
		*ppsfa = psfa;
	}
	return hr;
}


HRESULT CPrologQuery :: HrConvertSafeArrayElementToTerm (
	VARTYPE vt,
	void * pv,
	term_t tElem )
{
	HRESULT hr = S_OK;
	IP_IDispatch ipId;

	switch ( vt )
	{
		//  Leave "empty" variants as unbound terms
		case VT_NULL:
		case VT_EMPTY:
			PL_put_variable( tElem );
			break;

		//  Two-byte integers
		case VT_UI2:
		case VT_I2:
		{
			int i = *((SHORT*) pv);
			PL_put_integer( tElem, i );
			break;
		}

		case VT_INT:
		case VT_UI4:
		case VT_UINT:
		case VT_I4:
		{
			int i = *((LONG*) pv);
			PL_put_integer( tElem, i );
			break;
		}

		case VT_R4:
		{
			double d = *((float*)pv);
			PL_put_float( tElem, d );
			break;
		}

		case VT_R8:
		{
			double d = *((double*)pv);
			PL_put_float( tElem, d );
			break;
		}

		case VT_BSTR:
		{
			ZSTR zs( *((BSTR*) pv) );
			PL_put_string_chars( tElem, zs );
			break;
		}

		case VT_BOOL:
		{
			atom_t atm = *((SHORT*)pv) == VARIANT_TRUE
					 ? atomTrue
					 : atomFalse;
			PL_put_atom( tElem, atm );
			break;
		}

		case VT_DISPATCH:
		{
			ipId = *((IDispatch **) pv);
			if ( ! ipId )
				hr = Error( IDS_ERR_INV_OBJ_REF );
			break;
		}

		case VT_VARIANT:
		{
			VARIANT * pvnt = (VARIANT*) pv;
			vt = pvnt->vt;
			void * pv2 = & pvnt->byref;
			hr = HrConvertSafeArrayElementToTerm( vt, pv2, tElem );
			break;
		}

		default:
			hr = Error(IDS_ERR_SFA_DATA_FAILURE);
			break;
	}

	return hr;
}

HRESULT CPrologQuery :: HrSafeArrayBounds ( SAFEARRAY * psfa, int iDim, LONG & iLbound, LONG & iUbound )
{
	HRESULT hr = ::SafeArrayGetLBound( psfa, iDim, & iLbound );
	if ( FAILED(hr) )
		return hr;
	return ::SafeArrayGetUBound( psfa, iDim, & iUbound );
}

//
//	Convert a (guaranteeed) one-dimensional SAFEARRAY to Prolog terms.
//		Calling routine must have allocated sufficient terms.
//		
HRESULT CPrologQuery ::	HrConvertSafeArrayToTerms ( 
	VARTYPE vt,
	SAFEARRAY * psfa,
	term_t tArgs )
{
	assert( psfa && ::SafeArrayGetDim( psfa ) == 1 );
		
	LONG iLbound, iUbound;
	HRESULT hr = S_OK;
	
	int iDim = 1;
	VLONG vidim;
	vidim.resize(1);
	vidim[0] = 0;

	int cbElem = ::SafeArrayGetElemsize( psfa );	
	void * pv = _alloca( cbElem ); 
	do 
	{
		
		hr = HrSafeArrayBounds( psfa, iDim, iLbound, iUbound );
		if ( FAILED(hr) )
			break;

		for ( int iItem = iLbound; iItem <= iUbound; iItem++ )
		{
			term_t tElem = tArgs+iItem;
			vidim[0] = iItem;

			//  Yes- start extracting and converting elements
			hr = ::SafeArrayGetElement( psfa, & vidim[0], pv );
			if ( FAILED(hr) )
				break;

			hr = HrConvertSafeArrayElementToTerm( vt, pv, tElem );
			if ( FAILED(hr) )
				break;
		}
		if ( iItem < iUbound )
			break;
	}
	while (false);

	return hr;
}

STDMETHODIMP CPrologQuery::SetPredicate (
	BSTR bstrName, 
	long iArity, 
	BSTR bstrModule)
{
	ZSTR zsName( bstrName );
	ZSTR zsModule( bstrModule );

	ResetAll();

	CHDIR chdir( _zsModulePath );
			
	_iArity = iArity;
	_pred = PL_predicate( zsName, iArity, zsModule );
	return S_OK;
}

STDMETHODIMP CPrologQuery::OpenQuery (
	/*[in]*/ long iFlags, 
	/*[in]*/ SAFEARRAY * * prgArgs )
{
	CHDIR chdir( _zsModulePath );

	ResetQuery();

	SAFEARRAY * psfa = *prgArgs;
	assert( psfa );

	long iLbound, iUbound;
	HRESULT hr = S_OK;
	do
	{
		if ( _pred == NULL )
		{
			hr = Error( IDS_ERR_INV_PRED );
			break;
		}

		if ( _iArity != 0 )
		{
			if ( ::SafeArrayGetDim( psfa ) != 1 )	
			{
				hr = Error( IDS_ERR_INV_SFA_BOUNDS );
				break;
			}
			hr = HrSafeArrayBounds( psfa, 1, iLbound, iUbound );
			if ( FAILED(hr) )
				break;
			
			int iArity = iUbound + 1 - iLbound;
			if ( _iArity != iArity )
			{
				hr = Error(IDS_ERR_BAD_ARITY);
				break;
			}

			_tArgs = PL_new_term_refs( _iArity );
	
			hr = HrConvertSafeArrayToTerms( VT_VARIANT, psfa, _tArgs );
			if ( FAILED( hr ) )
				break;
		}

		_qid = PL_open_query( 0, PL_Q_NODEBUG, _pred, _tArgs );
	}
	while (false);
	return hr;
}


STDMETHODIMP CPrologQuery :: NextSolution (	
	/*[out]*/ SAFEARRAY * * prgArgs,
	/*[out,retval]*/ VARIANT_BOOL * pbSucceed ) 
{
	CHDIR chdir( _zsModulePath );

	*pbSucceed = VARIANT_FALSE;

	if ( PL_next_solution( _qid ) )
	{
		*pbSucceed = VARIANT_TRUE;

		if ( _iArity > 0 )
		{
			HRESULT hr = HrConvertTermsToSafeArray( _tArgs, _iArity, prgArgs );
			if ( FAILED(hr) )
				return hr;
		}
	}
	return S_OK;
}

STDMETHODIMP CPrologQuery :: CutQuery ()
{
	// BUGBUG: temporarily the same as CloseQuery()
	return CloseQuery();
}

STDMETHODIMP CPrologQuery :: CloseQuery ()
{
	CHDIR chdir( _zsModulePath );

	ResetQuery();
	return S_OK;
}

