/*
 **************************************************************************
 *		SWI-Prolog ActiveX Extensions.
 *
 *	This software is provided in source and binary form free of charge to 
 *  users of SWI-Prolog.  No warranty of any kind is expressed or implied;
 *  in particular, this is entirely unsupported software.
 *
 *  Documentation for this product is available from the SWI-Prolog 
 *  web site (http://www.swi.psy.uva.nl/projects/SWI-Prolog/).
 *
 *	Version 1.0.1.
 *
 *	October 1, 1999
 *
 **************************************************************************
 */

//
//	BASICS.H:  Basic STL-based declarations
//
#pragma once

// disable "string too long - truncated to 255 characters in the debug information"
#pragma warning( disable : 4786 )

// disable "'this' : used in base member initializer list"
#pragma warning ( disable : 4355 )

//BUGBUG: Caused by 'valarray': revisit these.
// disable "unsafe use of type 'bool' in operation"
#pragma warning ( disable : 4804 )	
// disable "forcing value to bool 'true' or 'false'"
#pragma warning ( disable : 4800 )	

// disable "identifier truncated"
#pragma warning ( disable : 4786 )	

#define USE_STD_NAMESPACE using namespace std

// CRTP inclusions
#include <assert.h>

// STL Inclusions
#include <exception>
#include <vector>

USE_STD_NAMESPACE;

//  Handle conflicting Unicode defs
#if defined(_UNICODE) && !defined(UNICODE)
	#define UNICODE
#endif
#if defined(UNICODE) && !defined(_UNICODE)
	#define _UNICODE
#endif
#if defined(_UNICODE)
	typedef wchar_t TCHAR;
#else
	typedef char TCHAR;
#endif

//	General common typedefs

typedef TCHAR * TSZ;
typedef const TCHAR * TSZC;
typedef char	CHAR;
typedef CHAR * SZ;
typedef const CHAR * SZC;
typedef	SZC		SZCPRM;
typedef SZC		SZCVAL;
typedef wchar_t WCHAR;
typedef WCHAR * WSZ;
typedef const WCHAR * WSZC;
typedef unsigned int UINT;
typedef unsigned short USINT;
typedef unsigned long ULONG;
typedef long	LONG;
typedef int		INT;
typedef short	SINT;
typedef double	REAL;
typedef double	DBL;
typedef UINT	TOKEN;		// Parser token
typedef int		BOOL;		// Must remain int, because windows.h defines it also
typedef short	SHORT;
							
#ifndef VOID
#define VOID	void		// BUGBUG: Archaic usage
#endif


//  'qsort' interface function prototypedef
//typedef	INT		(*PFNCMP)(const VOID*, const VOID*);


//  Define common vector classes and macros to generalize declarations.

typedef vector<bool> vbool;		//  Vector of 'bool': lower case to distinguish from BOOL (in windows.h)

#define DEFINEV(T)		typedef vector<T> V##T
#define DEFINEVP(T)		typedef vector<T *> VP##T
#define DEFINEVCP(T)	typedef vector<const T *> VCP##T

DEFINEV(UINT);	
DEFINEV(VUINT);
DEFINEV(INT);
DEFINEV(USINT);
DEFINEV(REAL);
DEFINEV(DBL);
DEFINEV(VDBL);
DEFINEV(VVDBL);
DEFINEV(BOOL);
DEFINEV(VBOOL);
DEFINEV(TSZ);
DEFINEV(VTSZ);
DEFINEV(TSZC);
DEFINEV(VTSZC);
DEFINEV(SZ);
DEFINEV(VSZ);
DEFINEV(SZC);
DEFINEV(VSZC);

//
//	UBOUND: macro to return the number of elements in a static array
//
#ifndef UBOUND
  #define UBOUND(rg)  (sizeof rg/sizeof rg[0])
#endif

//	self
//  unconst_this
//	unconst_self
//
// A useful alias in member functions
#define self (*this)
#define unconst_this(p,T) (const_cast<T *> (p))
#define unconst_self(t,T) (const_cast<T &> (t))
